/* Id: //guest/paul_krause/perforce/integrations/jdev/src/com/perforce/jdev/P4FileSystem.java $
 * Copyright 2000-2002 SoundBite Corporation.   All Rights Reserved.
 */
package com.perforce.jdev;
import com.perforce.api.Client;
import com.perforce.api.Env;
import com.perforce.api.PerforceException;

import com.perforce.jdev.cm.P4ConnectionIterator;

import java.io.IOException;

import java.util.Iterator;
import java.util.List;

import oracle.ide.scm.SCMFile;
import oracle.ide.scm.SCMFileStatus;
import oracle.ide.scm.SCMFileSystem;
import oracle.ide.scm.SCMOptions;

import oracle.ide.scm.error.SCMException;
import oracle.ide.scm.error.SCMFileSystemException;

import oracle.jdeveloper.cm.CMException;

/**
 * Support for file operations that interact with Perforce.
 *
 * @author Paul Krause
 * @version $Revision: #7 $ $Date: 2002/11/08 $
 */
public class P4FileSystem implements SCMFileSystem {

   /**
    * Opens the files for editing.
    *
    * @param files some Files to open
    * @exception SCMException if an exception was raised
    * @deprecated since JDeveloper 9.0.3, use {@link #edit} instead
    */
   public void setWritable(List files)
      throws SCMException
   {
      final Iterator fili = files.iterator();
      while (fili.hasNext()) {
         final SCMFile f = (SCMFile) fili.next();
         final P4File pf = (P4File) f.getLatestStatus();
         if (pf.isOpen()) continue;
         pf.openForEdit(null);
      }
   }


   /**
    * Opens a file for deletion, terminating its branch.
    *
    * @param f File to delete
    * @exception SCMException if deletion failed
    */
   public void delete(SCMFile f)
      throws SCMException
   {
      try {
         final P4File p = (P4File) f.getStatus();
         p.openForDelete();
      } catch (IOException e) {
         final String msg = "p4 failed to delete file " + f;
         throw new SCMException(msg, e);
      } catch (PerforceException e) {
         final String msg = "p4 failed to delete file " + f;
         throw new SCMException(msg, e);
      }
   }

   /**
    * Branch a file to a new location.
    *
    * @param src Description of Parameter
    * @param dest Description of Parameter
    * @exception SCMException Description of Exception
    */
   public void copy(SCMFile src, SCMFile dest)
      throws SCMException
   {
      final P4File s4 = (P4File) src.getStatus();
      s4.integrate(dest, s4.getChange());
   }

   /**
    * Branches a file and terminiates the old branch in one change.
    *
    * @param src Original location
    * @param dest Replacement location
    * @exception SCMException if an I/O error occurs
    */
   public void move(SCMFile src, SCMFile dest)
      throws SCMException
   {
      final P4File s4 = (P4File) src.getStatus();
      s4.integrate(dest, s4.getChange());
      try {
         s4.openForDelete(false, false, null);
      } catch (IOException e) {
         throw new SCMException(e.getMessage(), e);
      } catch (PerforceException e) {
         throw new SCMException(e.getMessage(), e);
      }

   }

   /**
    * Add is supported (partly).
    *
    * @return <code>true</code>
    * @since JDeveloper 9.0.3
    */
   public boolean isAddSupported() {
      return true;
   }

   /**
    * Opens files for addition to a depot.
    * Has to guess which depot to use.
    *
    * @param sf <code>SCMFile[]</code> files to add
    * @exception SCMException if an error occurs
    * @since JDeveloper 9.0.3
    */
   public void add(SCMFile[] sf) throws SCMException {
      for (int i = 0; i < sf.length; i++) {
         final SCMFileStatus s = sf[i].getStatus();
         final P4File p4 = (s instanceof P4File) ? (P4File) s :
            new P4File(sf[i], guessEnv(sf[i]));
         p4.openForAdd();
      }
   }

   final Env guessEnv(SCMFile sf) throws SCMException {
      final Client c = guessClient(sf);
      if (c == null) throw new SCMFileSystemException("No client for " + sf);
      return c.getEnv();
   }

   
   final Client guessClient(SCMFile sf) throws SCMException {
      final String path = sf.getPath();
      Client best = null;
      try {
         final P4ConnectionIterator ci = new P4ConnectionIterator();
         while (ci.hasNext()) {
            final Client c = (Client) ci.next(Client.class.getName());
            if (c.getOwner().length() == 0) {
               try {
                  c.getEnv().checkValidity();
               } catch (PerforceException e) {
                  P4Client.log(c.getName() + ": " + e.getMessage());
                  continue; // with next client
               }
               c.sync();
               c.invalidate();
            }
            final String root = c.getRoot();
            if (!path.startsWith(root)) continue;
            if (best == null || root.length() > best.getRoot().length()) {
               best = c;
            }
         }
      } catch (IOException e) {
         final String msg = "I/O failure establishing client for {0}";
         throw new SCMException(msg, new Object[] {path}, e);
      } catch (CMException e) {
         final String msg = "IDE failure establishing client for {0}";
         throw new SCMException(msg, new Object[] {path}, e);
      }
      return best;
   }

   /**
    * Edit is supported.
    *
    * @return <code>true</code> 
    * @since JDeveloper 9.0.3
    */
   public boolean isEditSupported() {
      return true;
   }

   /**
    * Opens files for editing.
    *
    * @param sf <code>SCMFile[]</code> files to edit
    * @exception SCMException if an error occurs
    * @since JDeveloper 9.0.3
    */
   public void edit(SCMFile[] sf) throws SCMException {
      final SCMOptions opts = null;
      for (int i = 0; i < sf.length; i++) {
         final SCMFileStatus s = sf[i].getStatus();
         if (s instanceof P4File) {
            ((P4File) s).openForEdit(opts);
            continue;
         }
         final String msg = sf[i] + " is not controlled by Perforce";
         throw new SCMFileSystemException(msg);
      }
   }

   /**
    * Move is supported.
    *
    * @return <code>true</code>
    * @since JDeveloper 9.0.3
    */
   public boolean isMoveSupported() {
      return true;
   }

   /**
    * Moves files in the depot.
    *
    * @param src orginal <code>SCMFile[]</code> values
    * @param dest target <code>SCMFile[]</code> values
    * @exception SCMException if an error occurs
    * @since JDeveloper 9.0.3
    */
   public void move(SCMFile[] src, SCMFile[] dest) throws SCMException {
      for (int i = 0; i < src.length; i++) {
         move(src[i], dest[i]);
      }
   }

   /**
    * Remove is supported.
    *
    * @return <code>true</code>
    * @since JDeveloper 9.0.3
    */
   public boolean isRemoveSupported() {
      return true;
   }

   /**
    * Removes files from depot.
    *
    * @param sf <code>SCMFile[]</code> files to delete
    * @exception SCMException if an error occurs
    * @since JDeveloper 9.0.3
    */
   public void remove(SCMFile[] sf) throws SCMException {
      for (int i = 0; i < sf.length; i++) {
         delete(sf[i]);
      }
   }
}
