/* $Id: //guest/paul_krause/perforce/integrations/jdev/src/com/perforce/jdev/P4Node.java#4 $
 * Copyright 2000-2002 SoundBite Corporation.   All Rights Reserved.
 */
package com.perforce.jdev;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import javax.swing.Icon;

import oracle.ide.addin.Observer;
import oracle.ide.addin.UpdateMessage;
import oracle.ide.exception.ChangeVetoException;
import oracle.ide.model.Attributes;
import oracle.ide.model.DefaultAttributes;
import oracle.ide.model.Node;
import oracle.ide.scm.SCMFile;
import oracle.ide.scm.SCMFileStatus;
import oracle.ide.scm.SCMOptions;
import oracle.ide.scm.SCMSystem;
import oracle.ide.scm.env.SCMFileFactory;
import oracle.ide.scm.error.SCMException;
import oracle.ide.scm.util.SCMAssert;

import com.perforce.api.CommitException;
import com.perforce.api.PerforceException;
import com.perforce.api.Env;
import com.perforce.api.SourceControlObject;

/**
 * @author Paul Krause
 * @created Thu Apr 25 12:27:27 2002
 * @version $Revision: #4 $
 * @see oracle.ide.model.DefaultNode
 */
public abstract class P4Node implements SCMFile, SCMFileStatus, Node {

   /** P4Package delegate for the depot entry. */
   //private SourceControlObject p4Obj;

   /** Oracle SCM delegate for the client file. */
   private SCMFile scmFile;

   /** Is true if user has opened this node, and false if user has closed it. */
   private boolean openFlag;

   /** Observers of this node. */
   private Collection attachedObservers;

   /**
    * Creates a new <code>P4Node</code> instance.
    *
    * @param f a <code>SCMFile</code> value
    * @param sco a <code>SourceControlObject</code> value
    */
   protected P4Node(SCMFile f, SourceControlObject sco)
   {
      //p4Obj = sco;
      scmFile = f;
   }

   abstract protected SourceControlObject getP4Object();

   /**
    * A factory method to construct the approporiate P4Node for an
    * SCMFile.
    *
    * @param f a generic <code>SCMFile</code>
    * @param env a Perforce <code>Env</code> object
    * @return a new <code>P4Node</code> object
    */
   static P4Node create(SCMFile f, Env env)
   {
      if (f instanceof P4Node) {
         return (P4Node) f;
      }
      if (f.isProject()) {
         return new P4Project(f, env);
      }
      if (f.isDirectory()) {
         return new P4Dir(f, env);
      }
      return new P4File(f, env);
   }

   /**
    * Gets the value of the observers property
    *
    * @return The Observers value
    */
   private synchronized Collection getObservers()
   {
      if (attachedObservers != null) return attachedObservers;
      attachedObservers = new LinkedList();
      return attachedObservers;
   }

   /**
    * Gets the value of the observi property
    *
    * @return The Observi value
    */
   private synchronized Iterator getObservi()
   {
      return (attachedObservers == null) ? null :
         attachedObservers.iterator();
   }


   /**
    * Description of the Method
    *
    * @param opts a <code>SCMOptions</code> value
    * @exception SCMException Description of Exception
    */
   abstract void openForEdit(SCMOptions opts)
      throws SCMException;


   /**
    * Update the overlay of the SCM delegate.
    */
   void refreshOverlay() throws PerforceException {
      getP4Object().sync();
      SCMSystem.getEnvironment().getOverlays().updateOverlay(scmFile);
   }


   /**
    * Description of the Method
    *
    * @param msg Description of Parameter
    */
   void veto(final String msg)
   {
      final Exception e = new ChangeVetoException(msg);
      SCMAssert.fail(e);
   }

   /**
    * Sets the value of the comparator property
    *
    * @param p0 The new comparator value
    */
   public void setComparator(Comparator p0) { }

   /**
    * Gets the value of the icon property
    *
    * @return The Icon value
    */
   public Icon getIcon()
   {
      return scmFile.getIcon();
   }

   /**
    * Gets the value of the URL property
    *
    * @return The URL value
    */
   public URL getURL()
   {
      return scmFile.getURL();
   }

  /**
    * Sets the value of the URL property
    *
    * @param u The new URL value
    */
   public void setURL(URL u) {
      scmFile = createSCMFile(u);
   }

   private static SCMFile createSCMFile(URL u) {
      final SCMFileFactory util = SCMSystem.env().getFileFactory();
      return util.createFile(u.getPath());
   }

   /**
    * Gets the value of the timestamp property
    *
    * @return The Timestamp value
    */
   public long getTimestamp()
   {
      return getP4Object().getSyncTime();
   }

   /**
    * Gets the value of the "dirty bit".
    *
    * @return The Dirty value
    * @see #markDirty
    */
   public boolean isDirty() {
      return getP4Object().outOfSync(0);
   }

   /**
    * Gets the value of the controlled property.
    *
    * @return true if Controlled
    */
   public boolean isControlled() {
      return false;
   }

   /**
    * Gets the value of the directory property.
    *
    * @return true if a Directory
    */
   public boolean isDirectory() {
      return scmFile.isDirectory();
   }

   /**
    * Gets an interface to the cached status of this SCMFile.  P4Node
    * instances are self-reporting, but SMCFile objects generated by
    * the SCM subsystem generally are not, so this method is to be
    * used instead of a class cast.
    *
    * @return a self-updating status object
    */
   public SCMFileStatus getStatus() {
      return this;
   }

   /**
    * Gets file status.
    *
    * @return a <code>String</code> representaion of the SourceControlObject
    */
   public String getString() {
      return getP4Object().toString();
   }

   /**
    * Gets the name of the file, including extension, but not path.
    *
    * @return <code>getName() + getExtension();</code>
    */
   public String getFileName() {
      return scmFile.getFileName();
   }

   /**
    * Gets the core name of the file only, excluding path and extension.
    *
    * @return The Name value
    */
   public String getName()
   {
      return scmFile.getName();
   }

   /**
    * Gets the value of the extension property
    *
    * @return The Extension value (with dot)
    */
   public String getExtension() {
      return scmFile.getExtension();
   }

   /**
    * Gets the value of the parent property
    *
    * @return The Parent value
    */
   public SCMFile getParent() {
      final SCMFile p = scmFile.getParent();
      if (p == null) return null;
      if (p instanceof P4Node) return p;
      try {
         return (P4Node) p.getStatus();
      } catch (ClassCastException e) {
         final String msg = "Unable to cast SCMStatus to P4Node";
         P4Client.logException(this, "getParent", msg, e);
         return p;
      } catch (SCMException e) {
         final String msg = "SCM failure";
         P4Client.logException(this, "getParent", msg, e);
         return p;
      }
   }

   /**
    * Gets the value of the workspace property.
    *
    * @return true if Workspace
    */
   public boolean isWorkspace() {
      return scmFile.isWorkspace();
   }

   /**
    * Gets the value of the project property.
    *
    * @return true if Project
    */
   public boolean isProject() {
      return scmFile.isProject();
   }

   /**
    * Gets the value of the project sources property
    *
    * @return The ProjectSources value
    */
   public Iterator getProjectSources() {
      return scmFile.getProjectSources(); // XXX wrap Iterator? or generate my own SCMFile objects?
   }

   /**
    * Gets the value of the data property
    *
    * @return The Data value
    */
   public Object getData() {
      openFlag = true;
      return scmFile.getData(); // XXX get data from depot instead?
   }

   /**
    * Gets the value of the overlay icon property.
    *
    * @return The OverlayIcon value
    */
   public Icon getOverlayIcon() {
      // XXX consider pushing up all or part of method from P4File
      return null;
   }

   /**
    * Gets the value of the path property
    *
    * @return The Path value
    */
   public String getPath() {
      return scmFile.getPath();
   }

   /**
    * Gets the value of the open property
    *
    * @return The Open value
    */
   public boolean isOpen() {
      return openFlag;
   }


   /**
    * Gets the value of the new property
    *
    * @return false, Document exists in depot
    */
   public boolean isNew()
   {
      return false;
   }

   /**
    * Gets the value of the read only property
    *
    * @return true unless overridden
    */
   public boolean isReadOnly()
   {
      return true;
   }

   /**
    * Gets an iterator over any children of this Element.
    *
    * @return null as there are no children unless overridden.
    */
   public Iterator getChildren()
   {
      return null;
   }

   /**
    * Gets the value of the attributes property
    *
    * @return The Attributes value
    */
   public Attributes getAttributes() {
      return DefaultAttributes.EMPTY_ATTRIBUTES;
   }

   /**
    * Gets a boolean from <var>opts</var>.
    *
    * @param opts an <code>SCMOptions</code> value
    * @param name the option name
    * @param fallback a default value
    * @return a <code>boolean</code> value
    */
   public final boolean getBooleanOption (final SCMOptions opts,
                                          final String name,
                                          final boolean fallback)
   {
      if (opts == null) return fallback;
      final Boolean val = (Boolean) opts.getOption(this, name);
      if (val == null) return fallback;
      return val.booleanValue();
   }

   /**
    * Gets an object from <var>opts</var>.
    *
    * @param opts an <code>SCMOptions</code> value
    * @param name the option name
    * @param fallback a default value
    * @return an <code>Object</code> value
    */
   public final Object getOption (final SCMOptions opts,
                                  final String name,
                                  final Object fallback)
   {
      if (opts == null) return fallback;
      final Object val = opts.getOption(this, name);
      if (val == null) return fallback;
      return val;
   }


   /**
    * Gets a String from <var>opts</var>.
    *
    * @param opts an <code>SCMOptions</code> value
    * @param name the option name
    * @param fallback a default value
    * @return a <code>String</code> value
    */
   public final String getProperty (final SCMOptions opts,
                                    final String name,
                                    final String fallback)
   {
      return getOption(opts, name, fallback).toString();
   }

   /**
    * Identifies node category to IDE.
    *
    * @return an <code>int</code> assigned by IDE
    */
   public abstract int getCategory();


   /**
    * Gets the value of the short label property
    *
    * @return The ShortLabel value
    */
   public String getShortLabel()
   {
      return getFileName();
   }

   /**
    * Gets the value of the long label property
    *
    * @return The LongLabel value
    */
   public String getLongLabel()
   {
      return getP4Object().toString();
   }

   /**
    * Tests whether this Element is a Container.
    *
    * @return false, unless overriden by a Container
    */
   public boolean isLeaf()
   {
      return true;
   }

   /**
    * Gets the value of the child count property
    *
    * @return The ChildCount value
    */
   public int getChildCount()
   {
      return 0;
   }

   /**
    * Gets the value of the allows children property
    *
    * @return The AllowsChildren value
    */
   public boolean getAllowsChildren()
   {
      return false;
   }

   /**
    * Saves (commits) the contents of this Document to persistant
    * storage (to the depot).
    *
    * @param shallow if true, Containers should not commit children
    * @exception IOException if a CommitException occures
    */
   public void save(boolean shallow)
      throws IOException
   {
      try {
         getP4Object().commit();
      } catch (final CommitException e) {
         P4Client.logException(this, "save", "CommitException", e);
         throw new IOException() {
               public String getMessage() {
                  return e.getMessage();
               }
            };
      }
   }

   /**
    * Set or unset the "dirty bit."
    *
    * @param bit true for dirty, false for clean
    * @see #isDirty
    */
   public void markDirty(boolean bit)
   {
      if (bit) {
         getP4Object().refreshUpdateTime();
      }
      else {
         getP4Object().inSync();
      }
   }

   /**
    * Deletes the client file.  Does not touch the depot file, if any,
    * this is the responsibity of subclasses.
    *
    * @exception SCMException if an error occurs
    */
   public void delete() throws SCMException {
      scmFile.delete();
   }

   /**
    * Asks whether this SCMFile exists on the client.
    *
    * @return a verdict on whether the file exists
    * @see #isControlled to ask whether file exists in depot
    */
   public boolean exists() {
      return scmFile.exists();
   }

   /**
    * Renames the client file.  Does not affect the depot file, if
    * any, subclasses are responsible for that.
    *
    * @param f an <code>SCMFile</code> value
    * @exception SCMException if an error occurs
    * @since JDev 9.0.3 Preview
    */
   public void renameTo(SCMFile f) throws SCMException {
      scmFile.renameTo(f);
   }

   /**
    * Copies the client file.  Does not affect the depot file, if
    * any, subclasses are responsible for that.
    *
    * @param f an <code>SCMFile</code> value
    * @exception SCMException if an error occurs
    * @since JDev 9.0.3 Preview
    */
   public void copy(SCMFile f) throws SCMException {
      scmFile.copy(f);
   }

   /**
    * Asks whether a file is the parent directory.
    *
    * @param f Prospective parent
    * @return true if f is the parent
    */
   public boolean hasParent(SCMFile f) {
      SCMFile p = getParent();
      if (p == null) return false;
      return (f == p) || f.getPath().equals(p.getPath());
   }

   /**
    * Convert to normal File.
    *
    * @return client file
    */
   public File toFile() {
      return scmFile.toFile();
   }

   /**
    * Opens the Document and loads any associated depot data into
    * internal data structures. If the implementation of this method
    * completes successfully, the LazyLoadable.isOpen() method should
    * then return true.
    *
    * @exception IOException to report a failure opening the Document
    */
   public void open()
      throws IOException
   {
      openFlag = true;
   }

   /**
    * Close this Document.
    *
    * @exception IOException not really
    */
   public void close()
      throws IOException
   {
      openFlag = false;
   }

   /**
    * Add an observer to this subject's observer collection.
    *
    * @param o The Observer
    * @see #detach
    */
   public void attach(Observer o) {
      getObservers().add(o);
   }

   /**
    * Remove an observer from this subject's observer collection.
    *
    * @param o The Observer
    * @see #attach
    */
   public void detach(Observer o) {
      getObservers().remove(o);
   }

   /**
    * Notify members of this subject's observer collection that a
    * change has occured.
    *
    * @param subject What changed
    * @param change How it changed
    */
   public void notifyObservers(Object subject, UpdateMessage change) {
      final Iterator observi = getObservi();
      if (observi != null) {
         while (observi.hasNext()) {
            Observer o = (Observer) observi.next();
            o.update(subject, change);
         }
      }
   }

   /**
    * Description of the Method
    *
    * @return Description of the Returned Value
    */
   public boolean mayHaveChildren()
   {
      return getAllowsChildren();
   }

   /**
    * Description of the Method
    *
    * @return Description of the Returned Value
    */
   public Enumeration breadthFirstEnumeration()
   {
      return null;
   }

   /**
    * Description of the Method
    *
    * @return Description of the Returned Value
    */
   public Enumeration children()
   {
      return null;
   }

}
